#
#   Author:  Conan Albrecht <ca&byu,edu>
#   License: Apache Open Source License
#


# pointer to our app config
# Django looks for this exact variable name
default_app_config = 'django_mako_plus.Config'


# the version
from .version import __version__


# the app config and registry
from .apps import Config
from .registry import register_dmp_app

# the router, middleware, and view function decorator
from .middleware import RequestInitMiddleware
from .router import route_request, view_function


# converter decorator
from .converter import parameter_converter


# the middleware and template
# the template engine
from .engine import MakoTemplates


# the exceptions
from .exceptions import BaseRedirectException
from .exceptions import RedirectException
from .exceptions import PermanentRedirectException
from .exceptions import JavascriptRedirectException
from .exceptions import InternalRedirectException


# filters and tags
from .filters import django_syntax, jinja2_syntax, alternate_syntax
from .templatetags.django_mako_plus import dmp_include

# the http responses
from .http import HttpResponseJavascriptRedirect


# the convenience functions
#
# Instead of these functions, consider using request.dmp.render() and request.dmp.render_to_string(),
# which are monkey-patched onto every DMP-enabled app at load time.  See the documentation
# for information on why we do this.
#
from .convenience import render_template
from .convenience import render_template_for_path
from .convenience import get_template
from .convenience import get_template_for_path
from .convenience import get_template_loader
from .convenience import get_template_loader_for_path


# the utilities
from .util import get_dmp_instance
from .util import merge_dicts


# the urls
# I'm specifically not including urls.py here because I want it imported
# as late as possible (after all the apps are set up).  Django will import it
# when it processes the project's urls.py file.


# html content shortcuts
from .provider import links
from .provider import template_links, template_obj_links
# html content providers
from .provider.base import BaseProvider
from .provider.compile import CompileProvider, CompileScssProvider, CompileLessProvider
from .provider.static_links import LinkProvider, CssLinkProvider, JsLinkProvider, JsContextProvider, jscontext
from .provider.webpack import WebpackCssLinkProvider, WebpackJsLinkProvider, WebpackJsCallProvider



from django.utils.encoding import force_text
from django.conf import settings

from ..util import merge_dicts, flatten

import os
import os.path



##############################################################
###   Static File Providers

class BaseProvider(object):
    '''
    Abstract base provider class.  An instance is tied to a template at runtime.
    '''

    default_options = {
        # the group this provider is part of.  this only matters when
        # the html page limits the providers that will be called with
        # ${ django_mako_plus.links(group="...") }
        'group': 'styles',
        # whether enabled (see "Dev vs. Prod" in the DMP docs)
        'enabled': True,
    }

    def __init__(self, app_config, template_file, options):
        '''
        Creates the provider.

        When settings.DEBUG=True, the provider is recreated per request.
        When settings.DEBUG=False, this constructor is run only once per server run.

        The fields created here deconstruct the location of the template.  Examples:

        Regular location: /homepage/templates/index.html
            self.app_config:            AppConfig for "homepage"
            self.app_config.path:       "/absolute/path/to/homepage/"
            self.template_file:         "index.html"
            self.template_subdir_parts: [ "templates" ]                  # subdir parts within the app
            self.template_name:         "index"
            self.template_ext:          ".html"
            self.template:              "index"                          # name within "homepage/templates/"

        Subdir location: /homepage/templates/mail/welcome.mail
            self.app_config:            AppConfig for "homepage"
            self.app_config.path:       "/absolute/path/to/homepage/"
            self.template_file:         "welcome.mail"
            self.template_subdir_parts: [ "templates", "mail" ]          # subdir parts within the app
            self.template_name:         "welcome"
            self.template_ext:          ".mail"
            self.template:              "mail/welcome"                   # name within "homepage/templates/"
        '''
        self.app_config = app_config
        self.template_subdir, self.template_file = os.path.split(template_file)
        self.template_name, self.template_ext = os.path.splitext(self.template_file)
        template_subdir_parts = os.path.normpath(self.template_subdir).split(os.path.sep)
        self.template = os.path.join(*flatten(template_subdir_parts[1:], self.template_name))
        self.options = merge_dicts(self.default_options, options)     # combined options dictionary

    @property
    def group(self):
        return self.options['group']

    def start(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run starts.
        Initialize values in the data dictionary here.
        '''
        pass

    def provide(self, provider_run, data):
        '''Called on *each* template's provider list in the chain - use provider_run.write() for content'''
        pass

    def finish(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run finishes
        Finalize values in the data dictionary here.
        '''
        pass



from django.http import HttpResponseNotAllowed

from ..util import log
from .base import Router
from .decorators import view_function
from .router_function import ViewFunctionRouter



class ClassBasedRouter(Router):
    '''
    Router for class-based views.

    A class-based router can have several endpoints, such as get(),
    post(), or put().  A ViewFunctionRouter is created for each
    of the endpoints that exist in the given class.  So this is really
    just a meta-router that reroutes to the appropriate ViewFunctionRouter
    within it.
    '''
    def __init__(self, module, instance):
        self.instance = instance
        self.endpoints = {}
        for mthd_name in instance.http_method_names:  # get parameters from the first http-based method (get, post, etc.)
            func = getattr(instance, mthd_name, None)
            if func is not None:
                # class-based methods don't explicitly have to be decorated with @view_function.
                # they don't need that security because, by subclassing View, we know they are endpoints.
                # but we still need the view_function decorator for the parameter conversion, so we wrap
                # if needed.
                if not view_function.is_decorated(func):
                    func = view_function(func)
                self.endpoints[mthd_name] = ViewFunctionRouter(module, func)


    def get_response(self, request, *args, **kwargs):
        endpoint = self.endpoints.get(request.method.lower())
        if endpoint is not None:
            return endpoint.get_response(request, **kwargs)
        log.info('Method Not Allowed (%s): %s', request.method, request.path, extra={'status_code': 405, 'request': request})
        return HttpResponseNotAllowed([ e.upper() for e in self.endpoints.keys() ])


    @property
    def name(self):
        return self.instance.__class__.__qualname__


    def message(self, request, descriptive=True):
        if descriptive:
            return 'class-based view function {}.{}.{}'.format(request.dmp.module, self.name, request.dmp.function)
        return '{}.{}.{}'.format(request.dmp.module, self.name, request.dmp.function)



from django.apps import AppConfig
from django.template import engines



class Config(AppConfig):
    name = 'django_mako_plus'
    label = 'django_mako_plus'
    verbose_name = 'Django Mako Plus Templating Engine'

    def ready(self):
        '''Called by Django when the app is ready for use.'''
        # ensure the DMP template engine has been loaded (the Django shell doesn't load it until it is accessed)
        # we just have to access the engine to cause the Django EngineHandler class to initialize it
        # Django calculates the alias/name based on its package, so it must be called django_mako_plus.
        # See the creation of EngineHandler.default_name in django.templates.util for this.
        engines['django_mako_plus']

        # trigger sorting of the converters now that the models are available
        from .converter.base import ParameterConverter
        ParameterConverter._sort_converters(app_ready=True)



from django.db.models import Model
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.http import HttpRequest

from .decorators import parameter_converter

import inspect
import datetime
import decimal



###   object (fallback if nothing else matches)  ###

@parameter_converter(object)
def convert_object(value, parameter):
    '''
    Fallback converter when nothing else matches:
        '', None convert to parameter default
        Anything else is returned as-is
    '''
    return _check_default(value, parameter, ( '', None ))


###  str (a passthrough)  ###

@parameter_converter(str)
def convert_str(value, parameter):
    '''
    Converts to string:
        '', None convert to parameter default
        Anything else is returned as-is (url params are already strings)
    '''
    return _check_default(value, parameter, ( '', None ))


###  int  ###

@parameter_converter(int)
def convert_int(value, parameter):
    '''
    Converts to int or float:
        '', '-', None convert to parameter default
        Anything else uses int() or float() constructor
    '''
    value = _check_default(value, parameter, ( '', '-', None ))
    if value is None or isinstance(value, int):
        return value
    try:
        return int(value)
    except Exception as e:
        raise ValueError(str(e))


###  float  ###

@parameter_converter(float)
def convert_float(value, parameter):
    '''
    Converts to int or float:
        '', '-', None convert to parameter default
        Anything else uses int() or float() constructor
    '''
    value = _check_default(value, parameter, ( '', '-', None ))
    if value is None or isinstance(value, float):
        return value
    try:
        return float(value)
    except Exception as e:
        raise ValueError(str(e))


###  decimal.Decimal  ###

@parameter_converter(decimal.Decimal)
def convert_decimal(value, parameter):
    '''
    Converts to decimal.Decimal:
        '', '-', None convert to parameter default
        Anything else uses Decimal constructor
    '''
    value = _check_default(value, parameter, ( '', '-', None ))
    if value is None or isinstance(value, decimal.Decimal):
        return value
    try:
        return decimal.Decimal(value)
    except Exception as e:
        raise ValueError(str(e))


###  bool  ###

@parameter_converter(bool)
def convert_boolean(value, parameter, default=False):
    '''
    Converts to boolean (only the first char of the value is used):
        '', '-', None convert to parameter default
        'f', 'F', '0', False always convert to False
        Anything else converts to True.
    '''
    value = _check_default(value, parameter, ( '', '-', None ))
    if isinstance(value, bool):
        return value
    if isinstance(value, str) and len(value) > 0:
        value = value[0]
    return value not in ( 'f', 'F', '0', False, None )


###  datetime.datetime  ###

@parameter_converter(datetime.datetime)
def convert_datetime(value, parameter):
    '''
    Converts to datetime.datetime:
        '', '-', None convert to parameter default
        The first matching format in settings.DATETIME_INPUT_FORMATS converts to datetime
    '''
    value = _check_default(value, parameter, ( '', '-', None ))
    if value is None or isinstance(value, datetime.datetime):
        return value
    for fmt in settings.DATETIME_INPUT_FORMATS:
        try:
            return datetime.datetime.strptime(value, fmt)
        except (ValueError, TypeError):
            continue
    raise ValueError("`{}` does not match a format in settings.DATETIME_INPUT_FORMATS".format(value))


###   datetime.date  ###

@parameter_converter(datetime.date)
def convert_date(value, parameter):
    '''
    Converts to datetime.date:
        '', '-', None convert to parameter default
        The first matching format in settings.DATE_INPUT_FORMATS converts to datetime
    '''
    value = _check_default(value, parameter, ( '', '-', None ))
    if value is None or isinstance(value, datetime.date):
        return value
    for fmt in settings.DATE_INPUT_FORMATS:
        try:
            return datetime.datetime.strptime(value, fmt).date()
        except (ValueError, TypeError):
            continue
    raise ValueError("`{}` does not match a format in settings.DATE_INPUT_FORMATS".format(value))


###   Model: any Django model by its id  ###

@parameter_converter(Model)  # django models.Model
def convert_id_to_model(value, parameter):
    '''
    Converts to a Model object.
        '', '-', '0', None convert to parameter default
        Anything else is assumed an object id and sent to `.get(id=value)`.
    '''
    value = _check_default(value, parameter, ( '', '-', '0', None ))
    if isinstance(value, (int, str)):  # only convert if we have the id
        try:
            return parameter.type.objects.get(id=value)
        except (MultipleObjectsReturned, ObjectDoesNotExist) as e:
            raise ValueError(str(e))
    return value



###################################
###   Helpers

def _check_default(value, parameter, default_chars):
    '''Returns the default if the value is "empty"'''
    # not using a set here because it fails when value is unhashable
    if value in default_chars:
        if parameter.default is inspect.Parameter.empty:
            raise ValueError('Value was empty, but no default value is given in view function for parameter: {} ({})'.format(parameter.position, parameter.name))
        return parameter.default
    return value



from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse, HttpResponseServerError, StreamingHttpResponse

from .factory import get_router
from .router_class import ClassBasedRouter
from .router_exception import RegistryExceptionRouter

from ..exceptions import InternalRedirectException, RedirectException
from ..registry import ensure_dmp_app
from ..signals import dmp_signal_internal_redirect_exception, dmp_signal_redirect_exception
from ..util import DMP_OPTIONS, get_dmp_instance, log, URLParamList

import sys

from urllib.parse import unquote




##############################################################
###   The front controller of all views on the site.
###   urls.py routes everything through this method.

def route_request(request, *args, **kwargs):
    '''
    The main router for all calls coming in to the system.  Patterns in urls.py should call this function.
    '''
    # an outer try that catches the redirect exceptions
    try:
        # ensure the middleware ran
        if getattr(request, 'dmp', None) is None:
            raise ImproperlyConfigured("Variable request.dmp.function_obj does not exist (check MIDDLEWARE for `django_mako_plus.RequestInitMiddleware`).")

        # output the variables so the programmer can debug where this is routing
        log.info(str(request.dmp))

        # if we had a view not found, raise a 404
        if isinstance(request.dmp.function_obj, RegistryExceptionRouter):
            log.info(request.dmp.function_obj.message(request))
            return request.dmp.function_obj.get_response(request, *args, **kwargs)

        # call the function
        log.info('calling %s', request.dmp.function_obj.message(request))
        response = request.dmp.function_obj.get_response(request, *args, **kwargs)

        # if we didn't get a correct response back, send a 500
        if not isinstance(response, (HttpResponse, StreamingHttpResponse)):
            msg = '%s failed to return an HttpResponse (or the post-signal overwrote it).  Returning 500 error.' % request.dmp.function_obj.message(request)
            log.info(msg)
            return HttpResponseServerError(msg)

        # return the response
        return response

    except InternalRedirectException as ivr:
        # send the signal
        if DMP_OPTIONS['SIGNALS']:
            dmp_signal_internal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=ivr)
        # update the RoutingData object
        request.dmp.module = ivr.redirect_module
        request.dmp.function = ivr.redirect_function
        request.dmp._set_function_obj(request, get_router(request.dmp.module, request.dmp.function, verify_decorator=False))
        if isinstance(request.dmp.function_obj, RegistryExceptionRouter):
            log.info('could not fulfill InternalViewRedirect because %s.%s does not exist.', request.dmp.module, request.dmp.function)
        # recurse with the same info
        log.info('received an InternalViewRedirect to %s.%s', request.dmp.module, request.dmp.function)
        return route_request(request, *args, **kwargs)

    except RedirectException as e: # redirect to another page
        if request.dmp.class_obj is None:
            log.info('%s redirected processing to %s', request.dmp.function_obj.message(request), e.redirect_to)
        # send the signal
        if DMP_OPTIONS['SIGNALS']:
            dmp_signal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=e)
        # send the browser the redirect command
        return e.get_response(request)

    # the code should never get here




#########################################################
###  Routing data for DMP.  Attached as `request.dmp`
###  by the middleware.


class RoutingData(object):
    '''
    A default object is set at the beginning of the request.
    Then after the url pattern is matched, an object with the
    match data is created.
    '''
    def __init__(self, request=None, app=None, page=None, function=None, urlparams=None):
        '''These variables are set by the process_view method above'''
        self.request = request

        # period and dash cannot be in python names, but we allow dash in app, dash in page, and dash/period in function
        self.app = app.replace('-', '_') if app is not None else None
        self.page = page.replace('-', '_') if page is not None else None
        if function:
            self.function = function.replace('.', '_').replace('-', '_')
            fallback_template = '{}.{}.html'.format(page, function)
        else:
            self.function = 'process_request'
            fallback_template = '{}.html'.format(page)

        # set the module and function
        # the return of get_router_function might be a function, a class-based view, or a template
        if self.app is not None and self.page is not None:
            self.module = '.'.join([ self.app, 'views', self.page ])
            self._set_function_obj(self.request, get_router(self.module, self.function, self.app, fallback_template))
        else:
            self.module = None
            self.class_obj = None
            self.function_obj = None

        # parse the urlparams
        # note that I'm not using unquote_plus because the + switches to a space *after* the question mark (in the regular parameters)
        # in the normal url, spaces should be quoted with %20.  Thanks Rosie for the tip.
        if isinstance(urlparams, (list, tuple)):
            self.urlparams = URLParamList(urlparams)
        elif urlparams:
            self.urlparams = URLParamList(( unquote(s) for s in urlparams.split('/') ))
        else:
            self.urlparams = URLParamList()


    def _set_function_obj(self, request, function_obj):
        '''Sets the function object, with alterations if class-based routing'''
        self.class_obj = None
        self.function_obj = function_obj
        if isinstance(self.function_obj, ClassBasedRouter):
            self.class_obj = self.function_obj
            self.function = request.method.lower()


    def __repr__(self):
        return 'RoutingData: app={}, page={}, module={}, function={}, urlparams={}'.format(
             self.app,
             self.page,
             self.module if self.class_obj is None else (self.module + '.' + self.class_obj.name),
             self.function,
             self.urlparams,
        )

    def _debug(self):
        return 'django_mako_plus RoutingData:' + \
            ''.join(('\n\t{: <16}{}'.format(k, v) for k, v in (
                ( 'app', self.app ),
                ( 'page', self.page ),
                ( 'module', self.module ),
                ( 'function', self.function ),
                ( 'class_obj', self.class_obj ),
                ( 'function_obj', self.function_obj ),
                ( 'urlparams', self.urlparams ),
            )))


    def render(self, template, context=None, def_name=None, subdir='templates', content_type=None, status=None, charset=None):
        '''App-specific render function that renders templates in the *current app*, attached to the request for convenience'''
        ensure_dmp_app(self.app)
        template_loader = get_dmp_instance().get_template_loader(self.app, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render_to_response')(context=context, request=self.request, def_name=def_name, content_type=content_type, status=status, charset=charset)


    def render_to_string(self, template, context=None, def_name=None, subdir='templates'):
        '''App-specific render function that renders templates in the *current app*, attached to the request for convenience'''
        ensure_dmp_app(self.app)
        template_loader = get_dmp_instance().get_template_loader(self.app, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render')(context=context, request=self.request, def_name=def_name)



from django.conf import settings
from django.core.exceptions import ViewDoesNotExist
from django.template import TemplateDoesNotExist
from django.views.generic import View

from .router_class import ClassBasedRouter
from .router_exception import RegistryExceptionRouter
from .router_function import ViewFunctionRouter
from .router_template import TemplateViewRouter

from .decorators import view_function

import inspect
import threading
from importlib import import_module
from importlib.util import find_spec




########################################################
###   Cached routers

CACHED_ROUTERS = {}
rlock = threading.RLock()

def get_router(module_name, function_name, fallback_app=None, fallback_template=None, verify_decorator=True):
    '''
    Gets or creates a mini-router for module_name.function_name.
    If the module or function cannot be found, ViewDoesNotExist is raised.
    '''
    # first check the cache
    key = ( module_name, function_name )
    try:
        return CACHED_ROUTERS[key]
    except KeyError:
        with rlock:
            # try again now that we're locked
            try:
                return CACHED_ROUTERS[key]
            except KeyError:
                func = router_factory(module_name, function_name, fallback_app, fallback_template, verify_decorator)
                if not settings.DEBUG:  # only cache in production mode
                    CACHED_ROUTERS[key] = func
                return func

    # the code should never be able to get here
    raise Exception("Django-Mako-Plus error: registry.get_router() should not have been able to get to this point.  Please notify the owner of the DMP project.  Thanks.")



############################################################
###   Creates the appropriate router for a
###   view function, class-based view function,
###   template, or error.

def router_factory(module_name, function_name, fallback_app=None, fallback_template=None, verify_decorator=True):
    '''
    Factory method to create a view-specific router in the system.
    See the four mini-routers at the end of this file.
    '''
    try:
        # I'm first calling find_spec first here beacuse I don't want import_module in
        # a try/except -- there are lots of reasons that importing can fail, and I just want to
        # know whether the file actually exists.  find_spec raises AttributeError if not found.
        try:
            spec = find_spec(module_name)
        except ValueError:
            spec = None
        if spec is None:
            # no view module, can we call the template directly?
            try:
                return TemplateViewRouter(fallback_app, fallback_template)
            except TemplateDoesNotExist as e:
                raise ViewDoesNotExist('View module {} not found, and fallback template {} could not be loaded ({})'.format(module_name, fallback_template, e))

        # load the module and function
        try:
            module = import_module(module_name)
            func = getattr(module, function_name)
        except ImportError as e:
            raise ViewDoesNotExist('Module "{}" could not be imported: {}'.format(module_name, e))
        except AttributeError as e:
            raise ViewDoesNotExist('Module "{}" found successfully, but "{}" was not found: {}'.format(module_name, function_name, e))

        # class-based view?
        if inspect.isclass(func) and issubclass(func, View):
            # must do func() to instantiate because func is class (not a function)
            return ClassBasedRouter(module, func())

        # a view function?
        if verify_decorator and not view_function.is_decorated(func):
            raise ViewDoesNotExist("View {}.{} was found successfully, but it must be decorated with @view_function or be a subclass of django.views.generic.View.".format(module_name, function_name))
        return ViewFunctionRouter(module, func)

    except ViewDoesNotExist as vdne:
        return RegistryExceptionRouter(vdne)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from ..command import run_command
from ..util import merge_dicts
from .base import BaseProvider

import os
import os.path
import shutil
import collections


class CompileProvider(BaseProvider):
    '''
    Runs a command, such as compiling *.scss or *.less, when an output file
    timestamp is older than the source file. In production mode, this check
    is done only once (the first time a template is run) per server start.

    When settings.DEBUG=True, checks for a recompile every request.
    When settings.DEBUG=False, checks for a recompile only once per server run.
    '''
    default_options = merge_dicts(BaseProvider.default_options, {
        'group': 'styles',
        # the source filename to search for
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        # codes: {app}, {template}, {template_name}, {template_file}, {template_subdir}
        'sourcepath': os.path.join('styles', '{template}.scss'),
        # the destination filename to search for
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        # codes: {app}, {template}, {template_name}, {template_file}, {template_subdir}, {sourcepath}
        'targetpath': os.path.join('styles', '{template}.css'),
        # the command to be run, as a list (see subprocess module)
        # codes: {app}, {template}, {template_name}, {template_file}, {template_subdir}, {sourcepath}, {targetpath}
        'command': [ 'echo', 'Subclasses should override this option' ],
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # since this is in the constructor, it runs only one time per server
        # run when in production mode
        if self.needs_compile:
            run_command(*self.build_command())

    @property
    def source(self):
        '''
        The absolute path to the file on disk.  This default implementation uses:

            development: /app path/
            production:  settings.STATIC_ROOT/app name/ during production.
        '''
        if settings.DEBUG:
            return os.path.join(
                self.app_config.path,
                self.options['sourcepath'].format(
                    app=self.app_config.name,
                    template=self.template,
                    template_name=self.template_name,
                    template_file=self.template_file,
                    template_subdir=self.template_subdir,
                ),
            )
        else:
            return os.path.join(
                settings.STATIC_ROOT,
                self.app_config.name,
                self.options['sourcepath'].format(
                    app=self.app_config.name,
                    template=self.template,
                    template_name=self.template_name,
                    template_file=self.template_file,
                    template_subdir=self.template_subdir,
                ),
            )

    @property
    def target(self):
        if settings.DEBUG:
            return os.path.join(
                self.app_config.path,
                self.options['targetpath'].format(
                    app=self.app_config.name,
                    template=self.template,
                    template_name=self.template_name,
                    template_file=self.template_file,
                    template_subdir=self.template_subdir,
                    sourcepath=self.source,
                ),
            )
        else:
            return os.path.join(
                settings.STATIC_ROOT,
                self.app_config.name,
                self.options['targetpath'].format(
                    app=self.app_config.name,
                    template=self.template,
                    template_name=self.template_name,
                    template_file=self.template_file,
                    template_subdir=self.template_subdir,
                    sourcepath=self.source,
                ),
            )

    def build_command(self):
        '''Returns the command to run, as a list/tuple (see subprocess module)'''
        if not isinstance(self.options['command'], collections.Iterable) or isinstance(self.options['command'], (str, bytes)):
            raise ImproperlyConfigured('The `command` option on a compile provider must be a list')
        return [
            str(arg).format(
                app=self.app_config.name,
                template=self.template,
                template_name=self.template_name,
                template_file=self.template_file,
                template_subdir=self.template_subdir,
                sourcepath=self.source,
                targetpath=self.target,
            )
            for arg in self.options['command']
        ]

    @property
    def needs_compile(self):
        try:
            source_mtime = os.stat(self.source).st_mtime
        except OSError:  # no source for this template, so just return
            return False
        try:
            target_mtime = os.stat(self.target).st_mtime
        except OSError: # target doesn't exist, so compile
            return True
        # both source and target exist, so compile if source newer
        return source_mtime > target_mtime


###################
###   Sass

class CompileScssProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.scss files.'''
    default_options = merge_dicts(CompileProvider.default_options, {
        # the source filename to search for
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'sourcepath': os.path.join('styles', '{template}.scss'),
        # the destination filename to search for
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'targetpath': os.path.join('styles', '{template}.css'),
        # the command to be run, as a list (see subprocess module)
        # codes: {app}, {template}, {template_name}, {template_file}, {template_subdir}, {sourcepath}, {targetpath}
        'command': [
            shutil.which('sass'),
            '--load-path=.',
            '{sourcepath}',
            '{targetpath}',
        ],
    })




#####################
###   Less

class CompileLessProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.less files.'''
    default_options = merge_dicts(CompileProvider.default_options, {
        # the source filename to search for
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'sourcepath': os.path.join('styles', '{template}.less'),
        # the destination filename to search for
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'targetpath': os.path.join('styles', '{template}.css'),
        # the command to be run, as a list (see subprocess module)
        # codes: {app}, {template}, {template_name}, {template_file}, {template_subdir}, {sourcepath}, {targetpath}
        'command': [
            shutil.which('lessc'),
            '--source-map',
            '{sourcepath}',
            '{targetpath}',
        ],
    })




from ..util import merge_dicts
from .static_links import CssLinkProvider, JsLinkProvider
from .base import BaseProvider

import os
import os.path
import posixpath



class WebpackCssLinkProvider(CssLinkProvider):
    '''
    Generates a CSS <link> tag for the sitewide or app-level CSS bundle file, if it exists.
    '''
    default_options = merge_dicts(JsLinkProvider.default_options, {
        # the filename to search for (resolves to a single file, if it exists)
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'filepath': os.path.join('styles', '__bundle__.css'),
    })



class WebpackJsLinkProvider(JsLinkProvider):
    '''
    Generates a JS <script> tag for the sitewide or app-level JS bundle file, if it exists.
    In settings, this provider should be defined *before* WebpackJsCallProvider is defined.
    '''
    default_options = merge_dicts(JsLinkProvider.default_options, {
        # the filename to search for (resolves to a single file, if it exists)
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'filepath': os.path.join('scripts', '__bundle__.js'),
    })



class WebpackJsCallProvider(BaseProvider):
    '''
    Calls the app/page function(s) for the current template and its supertemplates.
    This provider is used when bundling page functions together into a single JS file.
    See the DMP docs regarding the use of webpack and __entry__.js files for more information.
    '''
    default_options = merge_dicts(BaseProvider.default_options, {
        'group': 'scripts',
    })


    def start(self, provider_run, data):
        data['funcs_to_run'] = []


    def provide(self, provider_run, data):
        # instead of adding a script tag for each template (like the JsLinkProider does),
        # we just need to trigger the function for this page from the function
        data['funcs_to_run'].append(posixpath.join(self.app_config.name, self.template))


    def finish(self, provider_run, data):
        '''Trigger the functions for the template inheritance.'''
        # using the `if` statement in the JS because the function will only
        # exist if a .js file actually existed in the scripts/ directory when
        # dmp_webpack command was run.
        provider_run.write('<script data-context="{contextid}">'.format(
            contextid=provider_run.uid,
        ))
        for template in data['funcs_to_run']:
            provider_run.write('if (DMP_CONTEXT.appBundles["%s"]) { DMP_CONTEXT.appBundles["%s"]() };' % (
                template,
                template
            ))
        provider_run.write('</script>')



from .util import log

import os, os.path
import sys
import subprocess
import time
from collections import namedtuple


################################################################
###   Run a shell command

ReturnInfo = namedtuple('CommandReturn', ( 'code', 'stdout', 'stderr' ))


def run_command(*args, raise_exception=True):
    '''
    Runs a command, piping all output to the DMP log.
    The args should be separate arguments so paths and subcommands can have spaces in them:

        ret = run_command('ls', '-l', '/Users/me/My Documents')
        print(ret.code)
        print(ret.stdout)
        print(ret.stderr)

    On Windows, the PATH is not followed.  This can be overcome with:

        import shutil
        run_command(shutil.which('program'), '-l', '/Users/me/My Documents')
    '''
    args = [ str(a) for a in args ]
    log.info('running %s', ' '.join(args))
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
    stdout, stderr = p.communicate()
    returninfo = ReturnInfo(p.returncode, stdout.decode('utf8'), stderr.decode('utf8'))
    if stdout:
        log.info('%s', returninfo.stdout)
    if raise_exception and returninfo.code != 0:
        raise CommandError(' '.join(args), returninfo)
    return returninfo


class CommandError(Exception):
    def __init__(self, command, returninfo):
        self.command = command
        self.returninfo = returninfo
        super().__init__('CommandError')

    def __str__(self):
        return '[return value: {}] {}; {}'.format(self.returninfo.code, self.returninfo.stdout[:1000], self.returninfo.stderr[:1000])



from .base import Router
from .router import route_request, RoutingData
from .factory import get_router
from .router_class import ClassBasedRouter
from .router_exception import RegistryExceptionRouter
from .router_function import ViewFunctionRouter
from .router_template import TemplateViewRouter
from .decorators import view_function



from ..util import get_dmp_instance
from .base import Router



class TemplateViewRouter(Router):
    '''Router for direct templates (used whe a view.py file doesn't exist but the .html does)'''
    def __init__(self, app_name, template_name):
        # not keeping the actual template objects because we need to get from the loader each time (Mako has its own cache)
        self.app_name = app_name
        self.template_name = template_name
        # check the template by loading it
        get_dmp_instance().get_template_loader(self.app_name).get_template(self.template_name)


    def get_response(self, request, *args, **kwargs):
        template = get_dmp_instance().get_template_loader(self.app_name).get_template(self.template_name)
        return template.render_to_response(request=request, context=kwargs)


    def message(self, request, descriptive=True):
        if descriptive:
            return 'template {} (view function {}.{} not found)'.format(self.template_name, request.dmp.module, request.dmp.function)
        return str(self.template_name)



from django.http import Http404
from .base import Router




class RegistryExceptionRouter(Router):
    '''Router for a registry exception (i.e. view not found).'''
    def __init__(self, exc):
        self.exc = exc


    def get_response(self, request, *args, **kwargs):
        raise Http404(str(self.exc))


    def message(self, request, descriptive=True):
        if descriptive:
            return 'RegistryExceptionRouter: {}'.format(self.exc)
        return str(self.exc)



from django.core.management import BaseCommand
from django.core.management.base import CommandParser
from django_mako_plus.util import get_dmp_path
from django_mako_plus.management.mixins import DMPCommandMixIn

from importlib import import_module
import os
import os.path
import sys
import pkgutil



SUBCOMMAND_DEST = 'subcommand'

class Command(DMPCommandMixIn, BaseCommand):
    '''Runs DMP subcommands'''
    help = "The dmp management command runs a number of subcommands.  See below for the available subcommands."
    # settings might not be available yet, so can't set locale or run checks
    # once a subcommand is selected, these are reset to the subcommand settings
    leave_locale_alone = True
    requires_system_checks = False

    def __init__(self, *args, **kwargs):
        self.saved_args = args
        self.saved_kwargs = kwargs
        self.parser = None
        super().__init__(*args, **kwargs)


    def fetch_subcommand(self, subcommand):
        mod = import_module('django_mako_plus.management.subcommands.{}'.format(subcommand))
        cmd = mod.Command(*self.saved_args, **self.saved_kwargs)
        cmd._called_from_command_line = self._called_from_command_line
        return cmd


    def add_arguments(self, parser):
        super().add_arguments(parser)
        self.parser = parser

        # add a subparser for each subcommand
        current_subcommand = None
        def create_subparser(**kwargs):
            # I have to ask Django to create the parser and then use that as
            # a parent (template) for a new parser because django doesn't allow
            # the extra args and kwargs in its factory method (lame).
            kwargs['parents'] = [ current_subcommand.create_parser('', '') ]
            return CommandParser(current_subcommand, **kwargs)
        subparsers = parser.add_subparsers(
            title='DMP subcommands: (`python3 manage.py dmp [subcommand] --help` for docs)',
            dest=SUBCOMMAND_DEST,
            metavar='',
            parser_class=create_subparser,
        )
        for name in find_subcommands():
            current_subcommand = self.fetch_subcommand(name)
            subparsers.add_parser(name, description=current_subcommand.help, help=current_subcommand.help, add_help=False)


    def handle(self, *args, **options):
        # the dmp_command is required (not sure how to set this in the arguments above?)
        if not options[SUBCOMMAND_DEST]:
            self.print_help('', self.parser.prog)
            return

        # trigger the subcommand handling
        subcommand = self.fetch_subcommand(options[SUBCOMMAND_DEST])
        self.leave_locale_alone = subcommand.leave_locale_alone
        self.requires_system_checks = subcommand.requires_system_checks
        subcommand.execute(*args, **options)



#########################################
###   Helper functions

def find_subcommands():
    # this is ripped from django.core.management.__init__.py and switched to our subcommand dir
    command_dir = os.path.join(get_dmp_path(), 'management', 'subcommands')
    return [name for _, name, is_pkg in pkgutil.iter_modules([command_dir])
            if not is_pkg and not name.startswith('_')]



from django.conf import settings
from django.utils.module_loading import import_string

from ..util import crc32, getdefaultattr, log, merge_dicts
from ..version import __version__
from .base import BaseProvider

import logging
import os
import os.path
import json


#####################################################
###   Abstract Link Provider

class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.
    '''

    # the default options are for CSS files
    default_options = merge_dicts(BaseProvider.default_options, {
        # the filename to search for (resolves to a single file, if it exists)
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        # codes: {app}, {template}, {template_name}, {template_file}, {template_subdir}
        'filepath': os.path.join('scripts', '{template}.js'),
        # if a template is rendered more than once in a request, should we link more than once?
        'skip_duplicates': False,
    })
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        path = self.filepath
        if log.isEnabledFor(logging.DEBUG):
            log.debug('[%s] %s searching for %s', self.template_file, self.__class__.__qualname__, path)
        # file time and version hash
        try:
            self.mtime = int(os.stat(path).st_mtime)
            if log.isEnabledFor(logging.INFO):
                log.info('[%s] found %s', self.template_file, path)
            # version_id combines current time and the CRC32 checksum of file bytes
            self.version_id = (self.mtime << 32) | crc32(path)
        except FileNotFoundError:
            self.mtime = 0
            self.version_id = 0


    @property
    def filepath(self):
        '''
        The absolute path to the file on disk.  This default implementation uses:

            development: /app path/
            production:  settings.STATIC_ROOT/app name/ during production.
        '''
        if settings.DEBUG:
            return os.path.join(
                self.app_config.path,
                self.options['filepath'].format(
                    app=self.app_config.name,
                    template=self.template,
                    template_name=self.template_name,
                    template_file=self.template_file,
                    template_subdir=self.template_subdir,
                ),
            )
        else:
            return os.path.join(
                settings.STATIC_ROOT,
                self.app_config.name,
                self.options['filepath'].format(
                    app=self.app_config.name,
                    template=self.template,
                    template_name=self.template_name,
                    template_file=self.template_file,
                    template_subdir=self.template_subdir,
                ),
            )


    def create_link(self, provider_run):
        '''
        If the file referenced in filepath() exists, this method is called
        to create the link to be included in the html.  Subclasses should
        provide an implementation for their type of file, such as
        <script> or <link>.
        '''
        return '<link rel="stylesheet" type="text/css" href="/web/path/to/file" />'


    def start(self, provider_run, data):
        # add a set to the request (fallback to provider_run if request is None) for skipping duplicates
        if self.options['skip_duplicates']:
            data['seen'] = getdefaultattr(
                provider_run.request.dmp if provider_run.request is not None else provider_run,
                '_LinkProvider_Filename_Cache_',
                factory=set,
            )
        # enabled providers in the chain go here
        data['enabled'] = []


    def provide(self, provider_run, data):
        filepath = self.filepath
        # delaying printing of tag to finish() because the JsContextProvider delays and this must go after it

        # short circuit if the file for this provider doesn't exist
        if self.mtime == 0:
            return
        # short circut if we're skipping duplicates and we've already seen this one
        if self.options['skip_duplicates']:
            if filepath in data['seen']:
                return
            data['seen'].add(filepath)
        # if we get here, this provider is enabled, so add it to the list
        data['enabled'].append(self)


    def finish(self, provider_run, data):
        for provider in data['enabled']:
            provider_run.write(provider.create_link(provider_run))



#############################################
###   Css and Js Link Providers


class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'styles',
        # the filename to search for (resolves to a single file, if it exists)
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'filepath': os.path.join('styles', '{template}.css'),
        # if a template is rendered more than once in a request, we usually don't
        # need to include the css again.
        'skip_duplicates': True,
    })

    def create_link(self, provider_run):
        '''Creates a link to the given URL'''
        if settings.DEBUG:
            relpath = os.path.relpath(self.filepath, settings.BASE_DIR)
        else:
            relpath = os.path.relpath(self.filepath, settings.STATIC_ROOT)
        return '<link data-context="{}" rel="stylesheet" type="text/css" href="{}{}?v={:x}" />'.format(
            provider_run.uid,
            settings.STATIC_URL,
            relpath.replace('\\', '/'),
            self.version_id,
        )


class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        # the filename to search for (resolves to a single file, if it exists)
        # if it does not start with a slash, it is relative to the app directory.
        # if it starts with a slash, it is an absolute path.
        'filepath': os.path.join('scripts', '{template}.js'),
        # if a template is rendered more than once in a request, we should link each one
        # so the script runs again each time the template runs
        'skip_duplicates': False,
        # whether to create an async script tag
        'async': False,
    })

    def create_link(self, provider_run):
        '''Creates a link to the given URL'''
        if settings.DEBUG:
            relpath = os.path.relpath(self.filepath, settings.BASE_DIR)
        else:
            relpath = os.path.relpath(self.filepath, settings.STATIC_ROOT)
        return '<script data-context="{}" src="{}{}?v={:x}"{}></script>'.format(
            provider_run.uid,
            settings.STATIC_URL,
            relpath.replace('\\', '/'),
            self.version_id,
            ' async="async"' if self.options['async'] else '',
        )



###################################
###  JS Context Provider

class JsContextProvider(BaseProvider):
    '''
    Adds all js_context() variables to DMP_CONTEXT.
    This should be listed before JsLinkProvider so the
    context variables are available during <script> runs.
    '''
    default_options = merge_dicts(LinkProvider.default_options, {
        # the group this provider is part of.  this only matters when
        # the html page limits the providers that will be called with
        # ${ django_mako_plus.links(group="...") }
        'group': 'scripts',
        # the encoder to use for the JSON structure
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = import_string(self.options['encoder'])
        self.template = "{}/{}".format(self.app_config.name, self.template_name)

    def start(self, provider_run, data):
        data['templates'] = []

    def provide(self, provider_run, data):
        data['templates'].append(self.template)

    def finish(self, provider_run, data):
        if len(data['templates']) == 0:
            return
        context_data = {
            jscontext('__router__'): {
                'template': self.template,
                'app': provider_run.request.dmp.app if provider_run.request is not None else None,
                'page': provider_run.request.dmp.page if provider_run.request is not None else None,
            },
        }
        for k in provider_run.context.kwargs:
            if isinstance(k, jscontext):
                value = provider_run.context[k]
                context_data[k] = value.__jscontext__() if hasattr(value, '__jscontext__') else value
        provider_run.write('<script>')
        provider_run.write('DMP_CONTEXT.set("{version}", "{contextid}", {data}, {templates});'.format(
            version=__version__,
            contextid=provider_run.uid,
            data=json.dumps(context_data, cls=self.encoder, separators=(',', ':')) if context_data else '{}',
            templates=json.dumps(data['templates']),
        ))
        provider_run.write('</script>')


class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.

    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity



from django.http import Http404
from django.core.exceptions import ImproperlyConfigured

from ..util import log
from .parameter import ViewParameter
from .info import ConverterFunctionInfo

import logging
import inspect
from operator import attrgetter



class ParameterConverter(object):
    '''
    Converts parameters using functions registered to types they convert.

    To create a converter for a new type, simply create a function
    that is decorated with @parameter_converter

    Customizing Parameter Conversion:

    The primary way to customize conversion is to create new @parameter_converter
    functions in your codebase.  This allows you to add new type converters.
    You can even create functions for types already handled by the built-in
    converters.  Your functions will override the built-in ones.

    If you need to customize more than just a few types, override:

        convert_value(): where individual parameters are converted.

        __call__(): the controller that iterates the parameter values
        and calls convert_value().
    '''

    # the registry of converters (populated by the @converter_function decorator)
    converters = []
    # this variable prevents sorting until Django is ready (because during sorting
    # we switch "myapp.MyModel" to the actual model instance)
    _sorting_enabled = False

    def __init__(self, view_function):
        self.view_function = view_function

        # inspect the parameters on the function
        # import pdb; pdb.set_trace()
        self.signature = inspect.signature(self.view_function)
        # not using typing.get_type_hints because it adds Optional() to None defaults, and we don't need to follow mro here
        param_types = getattr(self.view_function, '__annotations__', {})
        params = []
        for i, p in enumerate(self.signature.parameters.values()):
            params.append(ViewParameter(
                name=p.name,
                position=i,
                kind=p.kind,
                type=p.annotation, #param_types.get(p.name) or inspect.Parameter.empty,
                default=p.default,
            ))
        self.parameters = tuple(params)


    @classmethod
    def _register_converter(cls, conv_func, conv_type):
        '''Triggered by the @converter_function decorator'''
        cls.converters.append(ConverterFunctionInfo(conv_func, conv_type, len(cls.converters)))
        cls._sort_converters()


    @classmethod
    def _sort_converters(cls, app_ready=False):
        '''Sorts the converter functions'''
        # app_ready is True when called from DMP's AppConfig.ready()
        # we can't sort before then because models aren't ready
        cls._sorting_enabled = cls._sorting_enabled or app_ready
        if cls._sorting_enabled:
            for converter in cls.converters:
                converter.prepare_sort_key()
            cls.converters.sort(key=attrgetter('sort_key'))


    def convert_parameters(self, request, *args, **kwargs):
        '''
        Iterates the urlparams and converts them according to the
        type hints in the current view function.  This is the primary
        function of the class.
        '''
        args = list(args)
        urlparam_i = 0
        # add urlparams into the arguments and convert the values
        for parameter_i, parameter in enumerate(self.parameters):
            # skip request object, *args, **kwargs
            if parameter_i == 0 or parameter.kind is inspect.Parameter.VAR_POSITIONAL or parameter.kind is inspect.Parameter.VAR_KEYWORD:
                pass
            # value in kwargs?
            elif parameter.name in kwargs:
                kwargs[parameter.name] = self.convert_value(kwargs[parameter.name], parameter, request)
            # value in args?
            elif parameter_i - 1 < len(args):
                args[parameter_i - 1] = self.convert_value(args[parameter_i - 1], parameter, request)
            # urlparam value?
            elif urlparam_i < len(request.dmp.urlparams):
                kwargs[parameter.name] = self.convert_value(request.dmp.urlparams[urlparam_i], parameter, request)
                urlparam_i += 1
            # can we assign a default value?
            elif parameter.default is not inspect.Parameter.empty:
                kwargs[parameter.name] = self.convert_value(parameter.default, parameter, request)
            # fallback is None
            else:
                kwargs[parameter.name] = self.convert_value(None, parameter, request)
        return args, kwargs


    def convert_value(self, value, parameter, request):
        '''
        Converts a parameter value in the view function call.

            value:      value from request.dmp.urlparams to convert
                        The value will always be a string, even if empty '' (never None).

            parameter:  an instance of django_mako_plus.ViewParameter that holds this parameter's
                        name, type, position, etc.

            request:    the current request object.

        "converter functions" register with this class using the @parameter_converter
        decorator.  See converters.py for the built-in converters.

        This function goes through the list of registered converter functions,
        selects the most-specific one that matches the parameter.type, and
        calls it to convert the value.

        If the converter function raises a ValueError, it is caught and
        switched to an Http404 to tell the browser that the requested URL
        doesn't resolve to a page.

        Other useful exceptions that converter functions can raise are:

            RedirectException: redirects the browser (see DMP docs)
            InternalRedirectException: redirects processing internally (see DMP docs)
            Http404: returns a Django Http404 response
        '''
        try:
            # we don't convert anything without type hints
            if parameter.type is inspect.Parameter.empty:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug('skipping conversion of parameter `%s` because it has no type hint', parameter.name)
                return value

            # find the converter method for this type
            # I'm iterating through the list to find the most specific match first
            # The list is sorted by specificity so subclasses come before their superclasses
            for ci in self.converters:
                if issubclass(parameter.type, ci.convert_type):
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug('converting parameter `%s` using %s', parameter.name, ci.convert_func)
                    return ci.convert_func(value, parameter)

            # if we get here, there wasn't a converter or this type
            raise ImproperlyConfigured(message='No parameter converter exists for type: {}. Do you need to add an @parameter_converter function for the type?'.format(parameter.type))

        except ValueError as e:
            log.info('ValueError raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e)
            raise Http404('Parameter could not be converted')

        except Exception as e:
            log.info('Exception raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e)
            raise



####################################
###   The router base class

class Router(object):
    '''Common superclass of routers'''

    def get_response(self, request, *args, **kwargs):
        raise NotImplementedError()



from django.http import HttpResponse, StreamingHttpResponse
from django.core.exceptions import ImproperlyConfigured

from ..signals import dmp_signal_post_process_request, dmp_signal_pre_process_request
from ..util import DMP_OPTIONS, log, import_qualified

from .base import Router
from ..converter import ViewParameter

import inspect
import sys
import logging




########################################################################################
###   Router classes for single endpoints.  When a view is first accessed, one
###   of these "mini" routers is created for it and cached for future calls.


class ViewFunctionRouter(Router):
    '''
    Router for a view function.:
    '''
    def __init__(self, mod, func):
        self.module = mod
        self.function = func
        # create a parameter converter object specific to this function
        self.converter = None
        if DMP_OPTIONS['PARAMETER_CONVERTER'] is not None:
            try:
                self.converter = import_qualified(DMP_OPTIONS['PARAMETER_CONVERTER'])(self.function)
            except ImportError as e:
                raise ImproperlyConfigured('Cannot find PARAMETER_CONVERTER: {}'.format(str(e)))


    def get_response(self, request, *args, **kwargs):
        '''Converts urlparams, calls the view function, returns the response'''
        # convert the parameters
        if self.converter is not None:
            args, kwargs = self.converter.convert_parameters(request, *args, **kwargs)

        # send the pre-signal
        if DMP_OPTIONS['SIGNALS']:
            for receiver, ret_response in dmp_signal_pre_process_request.send(sender=sys.modules[__name__], request=request, view_args=args, view_kwargs=kwargs):
                if isinstance(ret_response, (HttpResponse, StreamingHttpResponse)):
                    return ret_response

        # call the view (which is really the view_function decorator, which converts and then calls the view)
        response = self.function(request, *args, **kwargs)

        # send the post-signal
        if DMP_OPTIONS['SIGNALS']:
            for receiver, ret_response in dmp_signal_post_process_request.send(sender=sys.modules[__name__], request=request, response=response, view_args=args, view_kwargs=kwargs):
                if ret_response is not None:
                    response = ret_response # sets it to the last non-None in the signal receiver chain

        # return the response
        return response


    def message(self, request, descriptive=True):
        if descriptive:
            return 'view function {}.{}'.format(request.dmp.module, request.dmp.function)
        return '{}.{}'.format(request.dmp.module, request.dmp.function)



from ..decorators import BaseDecorator
from ..converter.base import ParameterConverter
import inspect




##########################################
###   View-function decorator

class view_function(BaseDecorator):
    '''
    A decorator to signify which view functions are "callable" by web browsers.
    and to convert parameters using type hints, if provided.

    All endpoint functions, such as process_request, must be decorated as:

        @view_function
        function process_request(request):
            ...

    Or:

        @view_function(...)
        function process_request(request):
            ...
    '''
    # singleton set of decorated functions
    DECORATED_FUNCTIONS = set()

    def __init__(self, decorator_function, *args, **kwargs):
        '''Create a new wrapper around the decorated function'''
        super().__init__(decorator_function, *args, **kwargs)

        # flag the function as an endpoint. doing it on the actual function because
        # we don't know the order of decorators on the function. order only matters if
        # the other decorators don't use @wraps correctly .in that case, @view_function
        # will put DECORATED_KEY on the decorator function rather than the real function.
        # but even that is fine *as long as @view_function is listed first*.
        real_func = inspect.unwrap(decorator_function)
        self.DECORATED_FUNCTIONS.add(real_func)


    @classmethod
    def is_decorated(cls, f):
        '''Returns True if the given function is decorated with @view_function'''
        real_func = inspect.unwrap(f)
        return real_func in cls.DECORATED_FUNCTIONS



from django.conf import settings
from django_mako_plus import view_function, jscontext
from datetime import datetime, timezone

@view_function
def process_request(request):
    utc_time = datetime.utcnow()
    context = {
        # sent to index.html:
        'utc_time': utc_time,
        # sent to index.html and index.js:
        jscontext('utc_epoch'): utc_time.timestamp(),
    }
    return request.dmp.render('index.html', context)